import pytest
from pymaterial.materials import IsotropicMaterial


@pytest.mark.parametrize(
    "Em, nu, density, ndigits",
    [
        (2.0e5, 0.3, 1000, [3, 3, 3]),
    ],
)
def test_values(Em, nu, density, ndigits):
    material = IsotropicMaterial(Em, nu, density)
    compliance = material.get_compliance()
    assert len(compliance) == 6
    assert round(compliance[0, 0], ndigits[0]) == round(1 / Em, ndigits[0])
    assert round(compliance[0, 1], ndigits[1]) == round(-nu / Em, ndigits[1])
    assert round(compliance[3, 3], ndigits[2]) == round(2 * (1 + nu) / Em, ndigits[2])


def test_plane_stress_stiffness():
    material = IsotropicMaterial(2.0e5, 0.3, 1000)
    stiff = material.get_plane_stress_stiffness()
    assert stiff.shape == (3, 3)


def test_plane_strain_stiffness():
    material = IsotropicMaterial(2.0e5, 0.3, 1000)
    stiff = material.get_plane_strain_stiffness()
    assert stiff.shape == (3, 3)



import pytest
from pymaterial.failures import CuntzeFailure


@pytest.mark.parametrize(
    "Em, rs, stresses, strains, result, ndigits",
    [
        (1.0, [2.0, -2.0, 1.0, -1.0, 0.5], [0.0, 0.0, 0.0], [0, 0, 0], 0.0, 2),
    ],
)
def test_values(Em, rs, stresses, strains, result, ndigits):
    failure = CuntzeFailure(Em, rs[0], rs[1], rs[2], rs[3], rs[4])
    res = failure.get_failure(stresses, strains)
    assert round(res["cuntze"], ndigits) == result



from .ifailure import IFailure
from typing import Optional, List
from math import sqrt


class VonMisesFailure(IFailure):
    def __init__(self, yield_stress: float):
        """
        con Mises yield criterion
        Parameters
        ----------
        yield_stress: float
            equivalent tensile stress or equivalent von-Mises stress
        Examples
        --------
        Create a Von-Mises-yield criteria
        >>> criteria = VonMises(280)
        >>> crit_loading = [140.0, 0.0, 0.0]
        >>> criteria.get_failure(stresses=crit_loading)
        returns {``max-stress``: 0.5}
        >>> uncrit_loading = [140.0, 0.0, 0.0]
        >>> criteria.get_failure(stresses=crit_loading)
        returns {``max-stress``: 0.5}
        """

        if yield_stress <= 0:
            raise ValueError(
                f"Yield stress has to be greater 0! (recieved: {yield_stress})"
            )

        self.strength = yield_stress

    def get_failure(
        self,
        stresses: Optional[List[float]] = None,
        strains: Optional[List[float]] = None,
        temperature: Optional[float] = None,
    ):
        if stresses is None:
            raise ValueError("Requires a stress tensor in Voigt notation!")
        length = len(stresses)
        allowed_length = [3, 6]
        if length not in allowed_length:
            raise ValueError(
                f"Stresses has to be of length 3 (2D) or 6 (3d stress state). "
                f"Recieved stress vector of length {length}."
            )

        stress = 0

        s11 = stresses[0]
        s22 = stresses[1]

        if length == 3:
            s12 = stresses[2]
            stress = sqrt(s11**2 + s11 * s22 + s22**2 + 3 * s12)
        else:  # length == 6
            s33 = stresses[2]
            s31 = stresses[3]
            s23 = stresses[4]
            s12 = stresses[5]
            stress = sqrt(
                0.5 * ((s11 - s22) ** 2 + (s22 - s33) ** 2 + (s33 - s11) ** 2)
                + 3.0 * (s12**2 + s23**2 + s31**2)
            )

        return {"mises": stress / self.strength}



from typing import Optional, List


class IFailure:
    def get_failure(
        self,
        stresses: Optional[List[float]] = None,
        strains: Optional[List[float]] = None,
        temperature: Optional[float] = None,
    ) -> dict:
        """
        Computes the loading dependent failure value.
        Parameters
        ----------
        stresses : List[float], optional
            stress tensor in Voigt notation
        strains : List[float], optional
            strain tensor using the Voigt notation:
        temperature: List[float, optional
            Temperature in [K]
        Notes
        -----
        - **stress** tensor in Voigt notation:
          + 2D: [sigma_11, sigma_22, sigma_12]
          + 3D: [sigma_11, sigma_22, sigma_33, sigma_23, sigma_13, sigma_12]
        - **strain** tensor using the Voigt notation:
          + 2D: [eps_11, eps_22, gamma_12=2*eps_12]
          + 3D: [eps_11, eps_22, eps_33, gamma_12_23, gamma_12_13, gamma_12_12]
        - ****
        Returns
        -------
        dict
            Dictionary of failure id and value.
            Values larger 1.0 a equivalent to a failure of the material.
        """
        raise NotImplementedError



from .material import Material  # noqa
from .anisotropic import AnisotropicMaterial  # noqa
from .orthotropic import OrthotropicMaterial  # noqa
from .transversely_isotropic import TransverselyIsotropicMaterial  # noqa
from .isotropic import IsotropicMaterial  # noqa



from .ifailure import IFailure
from typing import Optional, List


class CuntzeFailure(IFailure):
    def __init__(
        self,
        E1: float,
        R_1t: float,
        R_1c: float,
        R_2t: float,
        R_2c: float,
        R_21: float,
        my_21: Optional[float] = 0.3,
        interaction: Optional[float] = 2.5,
    ):
        """
        Cuntze failure criterion [1]_

        Parameters
        ----------
        E1 : float
            E-Modulus in fibre direction
        R_1t : float
            tensile strength in fibre direction
        R_1c : float
            compressive strength in fibre direction
        R_2t : float
            tensile strength perpendicular to fibre direction
        R_2c : float
            compressive strength perpendicular to fibre direction
        R_21 : float
            in-plane shear strength
        my_21 : Optional[float], optional
            Reibungs-/Materialparameter 0 < My > 0.3, by default 0.3
        interaction : Optional[float], optional
            Interaktionsfaktor 2,5 < m < 3,1, by default 2.5

        References
        ---------
        .. [1] R.G. Cuntze and A. Freund,
            "The predictive capability of failure mode concept-based strength criteria
            for multidirectional laminates", Composites Science and Technology, vol. 64,
            no. 3, pp. 343-377, 2004
        """

        self.E1 = E1
        self.R_1t = R_1t
        self.R_1c = R_1c
        self.R_2t = R_2t
        self.R_2c = R_2c
        self.R_21 = R_21
        self.my_21 = my_21
        self.interaction = interaction

    def get_failure(
        self,
        stresses: Optional[List[float]] = None,
        strains: Optional[List[float]] = None,
        temperature: Optional[float] = None,
    ):
        # epsilon_xt tensile strain in fibre direction
        # epsilon_xc compressive strain in fibre direction
        # sigma_yt   tensile stress perpendicular to fibre direction
        # sigma_yc   compressive stress perpendicular to fibre direction
        # sigma_y    stress perpendicular to fibre direction
        # tau_yx     in-plane shear stress
        # my_21 (= 0.3)  # Reibungs-/Materialparameter 0 < My > 0.3
        m = self.interaction  # Interaktionsfaktor 2,5 < m < 3,1

        # Reserve factors for single layer according to Cuntze

        epsilon_x = strains[0]
        epsilon_xt = 0.0
        epsilon_xc = 0.0
        if epsilon_x > 0:
            epsilon_xt = epsilon_x
        else:
            epsilon_xc = epsilon_x

        sigma_y = stresses[1]
        sigma_yt = 0.0
        sigma_yc = 0.0
        if sigma_y > 0:
            sigma_yt = sigma_y
        else:
            sigma_yc = sigma_y

        tau_yx = stresses[2]

        eff_1sigma = (epsilon_xt * self.E1) / self.R_1t  # FF1
        eff_1tau = (abs(epsilon_xc) * self.E1) / self.R_1c  # FF2

        eff_2sigma = sigma_yt / self.R_2t  # IFF1
        eff_2tau = abs(sigma_yc) / self.R_2c  # IFF2
        eff_21 = abs(tau_yx) / (self.R_21 - self.my_21 * sigma_y)  # IFF2

        # total Reservefactor
        eff_ges = (
            eff_1sigma**m
            + eff_1tau**m
            + eff_2sigma**m
            + eff_2tau**m
            + eff_21**m
        ) ** (1 / m)

        return {"cuntze": eff_ges}



from typing import List, Tuple
import numpy as np
from .ply import Ply
from pymaterial.materials import TransverselyIsotropicMaterial


class Stackup:
    def __init__(self, plies: List[Ply], bot_to_top=True):
        """
        Stackup or Laminat
        Parameters
        ----------
        plies : List[Ply]
            list of lies that makes the stackup, order is bottom to top
        References
        ----------
        .. [1] J. Ashton and J.M. Whitney, "Theory of Laminated Plates",
           Technomic, vol. 4, 1970
        """

        self.plies = plies
        if not bot_to_top:
            self.plies.reverse()
            self.plies = plies
        self.thickness = self.calc_thickness()
        self.density = self.calc_density()
        self.abd = None

    def get_plies(self, bot_to_top=True) -> List[Ply]:
        """
        Get plies of stackup.
        Returns
        -------
        List[Ply]
            lists of plies, order is bottom to top
        """
        if not bot_to_top:
            copy = self.plies.copy()
            copy.reverse()
            return copy
        else:
            return self.plies

    def calc_thickness(self) -> float:
        """
        Calculate the stackup thickness
        Notes
        -----
        **Calculation** means that that the result will allways computated again.
        Meaning that it will always correct, but might take a little bit longer.
        If you dont want this use **get_thickness()** instead!
        Returns
        -------
        float
            complete thickness of the stackup or sum over all ply thicknesses
        """
        thickness = 0.0
        for ply in self.plies:
            thickness = thickness + ply.thickness
        self.thickness = thickness
        return self.thickness

    def get_thickness(self) -> float:
        """
        Returns the stackup thickness
        Notes
        -----
        The Thickness is calculated at the beginning of the object creation.
        Meaning that later changes will probably not be considered!
        If you dont want this use **calc_thickness()** instead!
        Returns
        -------
        float
            complete thickness of the stackup or sum over all ply thicknesses
        """
        return self.thickness

    def calc_density(self) -> float:
        """
        Calculate the stackup density
        Notes
        -----
        **Calculation** means that that the result will allways computated again.
        Meaning that it will always correct, but might take a little bit longer.
        If you dont want this use **get_density()** instead!
        Returns
        -------
        float
            density of the stackup or mean over the plies
        """
        thick_density = 0.0
        for i in range(len(self.plies)):
            ply = self.plies[i]
            density = ply.get_material().get_density()
            if density is None:
                raise ValueError(f"Density is not defined for Material in Layer {i}.")
            thick_density += ply.thickness * density
        self.density = thick_density / self.get_thickness()
        return self.density

    def get_density(self) -> float:
        """
        Returns the stackup density
        Notes
        -----
        The Density is calculated at the beginning of the object creation.
        Meaning that later changes will probably not be considered!
        If you dont want this use **calc_density()** instead!
        Returns
        -------
        float
            density of the stackup or mean over the plies
        """
        return self.density

    def rotate(self, angle, degree=False) -> "Stackup":
        """
        Rotate Stackup
        Parameters
        ----------
        angle
            angle in [rad] or [deg] when degree is set True
        degree : bool, optional
            changes the measurement system of angle, default is False
        Returns
        -------
        Stackup
            new instance of the rotated Stackup
        Examples
        --------
        >>> stackup = Stackup([ply])
        >>> rot_stackup = stackup.rotate(90, degree=True)
        Be aware, that the stackup is left untouched.
        To change this, use:
        >>> stackup = stackup.rotate(90, degree=True)
        instead.
        """
        if degree:
            angle = angle / 180.0 * np.pi
        plies = []
        for ply in self.plies:
            plies.append(ply.rotate(angle))
        return Stackup(plies)

    def get_abd(self, truncate=True) -> np.ndarray:
        """
        Returns the ABD-Matrix of the stackup.
        Parameters
        ----------
        truncate : bool, optional
            when true: erase values smaller 1e-6. default: True
        Returns
        -------
        array:
            ABD-Matrix, dim=(6,6)
        """
        if self.abd is None:
            h = self.thickness / 2

            # Create empty matricces for A B en D.
            A = np.zeros((3, 3))
            B = np.zeros((3, 3))
            D = np.zeros((3, 3))

            # Loop over all plies
            z_bot = -h
            for ply in self.plies:
                # Calculate the z coordinates of the top and bottom of the ply.
                z_top = z_bot + ply.thickness

                # Rotate the local stiffenss matrix.
                Q_bar = ply.get_stiffness()

                # Calculate the contribution to the A, B and D matrix of this layer.
                Ai = Q_bar * (z_top - z_bot)
                Bi = 1 / 2.0 * Q_bar * (z_top**2 - z_bot**2)
                Di = 1 / 3.0 * Q_bar * (z_top**3 - z_bot**3)

                # Summ this layer to the previous ones.
                A = A + Ai
                B = B + Bi
                D = D + Di
                z_bot = z_top

            # Compile the entirety of the ABD matrix.
            self.abd = np.array(
                [
                    [A[0, 0], A[0, 1], A[0, 2], B[0, 0], B[0, 1], B[0, 2]],
                    [A[1, 0], A[1, 1], A[1, 2], B[1, 0], B[1, 1], B[1, 2]],
                    [A[2, 0], A[2, 1], A[2, 2], B[2, 0], B[2, 1], B[2, 2]],
                    [B[0, 0], B[0, 1], B[0, 2], D[0, 0], D[0, 1], D[0, 2]],
                    [B[1, 0], B[1, 1], B[1, 2], D[1, 0], D[1, 1], D[1, 2]],
                    [B[2, 0], B[2, 1], B[2, 2], D[2, 0], D[2, 1], D[2, 2]],
                ]
            )

        # Truncate very small values.
        if truncate is True:
            return np.array(
                np.where(np.abs(self.abd) < np.max(self.abd) * 1e-6, 0, self.abd)
            )
        return self.abd

    def calc_homogenized(self) -> TransverselyIsotropicMaterial:
        """
        Homogenize the Stackup as a Transversely Isotropic Material.
        Returns
        -------
        TransverselyIsotropicMaterial
            Homogenized Stackup
        """
        scale = 1.0 / self.get_thickness()
        e_1 = scale * (
            self.get_abd()[0, 0] - (self.get_abd()[0, 1]) ** 2 / self.get_abd()[1, 1]
        )
        e_2 = scale * (
            self.get_abd()[1, 1] - (self.get_abd()[0, 1]) ** 2 / self.get_abd()[0, 0]
        )
        g_12 = scale * self.get_abd()[2, 2]
        nu_12 = self.get_abd()[0, 1] / self.get_abd()[1, 1]
        return TransverselyIsotropicMaterial(e_1, e_2, nu_12, g_12, self.get_density())

    def apply_load(self, mech_load: np.ndarray) -> np.ndarray:
        """
        Calculate the strain and curvature of the full plate
        under a given load using Kirchhoff plate theory.
        Parameters
        ----------
        mech_load : vector
            The load vector consits of are
            :math:`(N_x, N_y, N_{xy}, M_x, M_y, M_{xy})^T`
        Returns
        -------
        deformation : vector
            This deformation consists of :math:`(varepsilon_x, varepsilon_y
            varepsilon_{xy},kappa_x, kappa_y, kappa_{xy})^T`
        """
        return np.ravel(np.linalg.inv(self.get_abd()).dot(mech_load))

    def apply_deformation(self, deformation: np.ndarray) -> np.ndarray:
        """
        Calculate the applied load and moment of the plate
        under a given deformation using Kichhoff plate theory.
        Parameters
        ----------
        deformation : vector
            This deformation consists of :math:`(varepsilon_x, varepsilon_y,
            varepsilon_{xy},kappa_x, kappa_y, kappa_{xy})^T`
        Returns
        -------
        load : vector
            The load vector consits of are
            :math:`(N_x, N_y, N_{xy}, M_x, M_y, M_{xy})^T`
        """
        return np.ravel(self.get_abd().dot(deformation))

    def get_strains(
        self, deformation: np.ndarray
    ) -> List[Tuple[np.ndarray, np.ndarray]]:
        """
        Return strains in plies.
        Parameters
        ----------
        deformation : array
            deformation tensor of shape [e_11, e_11, e_12, x_11, x_22, x_12]
            with e being the membrane strain, and x = curvature
        Returns
        -------
        List[Tuple[np.ndarray, np.ndarray]]
            list of strains (bottom to top) with tuple of the lower and upper strain
        Notes
        -----
        Result can be used in get_stress(strains)
        """
        # Calculating total thickness of the layup.
        h = self.thickness / 2

        # Calculate deformation of the midplane of the laminate.
        strain_membrane = deformation[:3]
        curvature = deformation[3:]

        # Create a list for the strains in each ply.
        strains = []
        z_bot = -h

        # Iterate over all plies.
        for ply in self.plies:
            # Calculate the z coordinates of the top and bottom of the ply.
            z_top = z_bot + ply.thickness

            # Caluclate strain in the ply.
            strain_top = strain_membrane + z_top * curvature
            strain_bot = strain_membrane + z_bot * curvature

            # Rotate strain from global to ply axis sytstem.
            strain_lt_top = ply.get_local_strain(strain_top)
            strain_lt_bot = ply.get_local_strain(strain_bot)

            # Store the strain values of this ply.
            strain_ply = (strain_lt_bot, strain_lt_top)
            strains.append(strain_ply)
            z_bot = z_top

        return strains

    def get_stresses(
        self, strains: List[Tuple[np.ndarray, np.ndarray]]
    ) -> List[Tuple[np.ndarray, np.ndarray]]:
        """
        Return stresses in plies.
        Parameters
        ----------
        strains : list
            list of strains in plies
        Returns
        -------
        List[Tuple[np.ndarray, np.ndarray]]
            list of stresses (bottom to top) with tuple of the lower and upper stress
        Notes
        -----
        Result can be used in get_failure(stresses)
        """
        stresses = []

        # Iterate over all plies.
        for i in range(len(self.plies)):
            # Obtain the strains from this ply.
            strain_lt_bot, strain_lt_top = strains[i]

            # Convert strains into stresses.
            stiffness = self.plies[i].get_stiffness(local=True)
            stress_lt_top = stiffness.dot(strain_lt_top)
            stress_lt_bot = stiffness.dot(strain_lt_bot)

            # Store the stress values of this ply.
            stress_ply = (stress_lt_bot, stress_lt_top)
            stresses.append(stress_ply)

        return stresses

    def get_failure(
        self,
        stresses: List[Tuple[np.ndarray, np.ndarray]],
        strains: List[Tuple[np.ndarray, np.ndarray]],
    ) -> List[Tuple[dict, dict]]:
        """
        Return the failures in the plies.
        Parameters
        ----------
        strains : list
            list of strains in plies
        stresses : list
            list of stresses in plies
        Returns
        -------
        List[Tuple[dict, dict]]
            list of failures (bottom to top) with tuple of the lower and upper failure
        """
        failures = []
        for i in range(len(stresses)):
            ply_stress = stresses[i]
            ply_strains = strains[i]
            material = self.plies[i].get_material()
            failures.append(
                (
                    material.get_failure(
                        stresses=ply_stress[0], strains=ply_strains[0]
                    ),
                    material.get_failure(
                        stresses=ply_stress[1], strains=ply_strains[1]
                    ),
                )
            )
        return failures



from .ifailure import IFailure
from typing import Optional, List, Tuple, Union


class MaxStressFailure(IFailure):
    def __init__(self, stress_strength: List[Union[float, Tuple[float, float]]]):
        """
        Maximum-Stress Failure Criterion
        Parameters
        ----------
        stress_strength : List[Tuple[float, float]]
            strength tensor in Voigt notation
        Notes
        -----
        - strength tensor in Voigt notation with compression/min (comp)
          and tensile/max (tens) strength
          * 2D: [(comp_11, tens_11),
            (comp_22, tens_22),
            (comp_12, sigma_12)]
          * 3D: [(comp_11, tens_11),
            (comp_22, tens_22),
            (comp_33, tens_33),
            (comp_23, sigma_23),
            (comp_13, sigma_13),
            (comp_12, sigma_12)]
        - if **one** value is used instead of the tuple,
          then the is considerd the maximum value (so it should be positive)
          and the minimum will be assumed to be the negative version
        Examples
        --------
        Create a Plane-Maximum-Stress Failure Criteria
        >>> criteria = MaxStress([(0.0, 2.0), (-2.0, 2.0), 2.0])
        >>> crit_loading = [4.0, 2.0, 0.0]
        >>> criteria.get_failure(stresses=crit_loading)
        returns {``max-stress``: 2.0}
        >>> uncrit_loading = [1.0, 1.0, 1.0]
        >>> criteria.get_failure(stresses=crit_loading)
        returns {``max-stress``: 0.5}
        """

        self.stress_mapping = [0, 1, 2, 3, 4, 5]
        length = len(stress_strength)
        if length == 3:
            self.stress_mapping = [0, 1, 1, 2, 2, 2]
        elif length != 3 and length != 6:
            raise ValueError(
                f"Stress-Strength Tensor has to be size 3 or 6! (Got: {length})"
            )

        self.strength = []
        # generalize to s11, s22, s33, s23, s13, s12
        # note: sXY are tuples with (min, max)
        for i in range(6):
            strength = stress_strength[self.stress_mapping[i]]
            if not isinstance(strength, tuple):
                strength = (-strength, strength)
            self.strength.append(strength)

    def get_failure(
        self,
        stresses: Optional[List[float]] = None,
        strains: Optional[List[float]] = None,
        temperature: Optional[float] = None,
    ):
        if stresses is None:
            raise ValueError("Need stress tensor in Voigt notation!")
        length = len(stresses)
        allowed_length = max(self.stress_mapping) + 1
        if length != allowed_length:
            raise ValueError(
                f"Stresses has to be of length  {allowed_length} "
                f"({int(allowed_length / 3 + 1)}d stress), "
                f"but got length {length}."
            )

        load = []
        for i in range(6):
            load.append(stresses[self.stress_mapping[i]])

        factor = []
        for i in range(6):
            s_min, s_max = self.strength[i]
            middle = (s_max + s_min) / 2
            dist = (s_max - s_min) / 2
            factor.append(abs(load[i] - middle) / dist)
        return {"max-stress": max(factor)}



from .orthotropic import OrthotropicMaterial, ndarray, np, IFailure, Optional, List


class TransverselyIsotropicMaterial(OrthotropicMaterial):
    def __init__(
        self,
        E_l: float,
        E_t: float,
        nu_lt: float,
        G_lt: float,
        density: float,
        nu_tt: Optional[float] = None,
        failures: Optional[List[IFailure]] = None,
    ):
        self.E_l = E_l
        self.E_t = E_t
        self.nu_lt = nu_lt
        if nu_tt is None:
            self.nu_tt = nu_lt
        self.G_lt = G_lt
        super().__init__(
            E_x=E_l,
            E_y=E_t,
            E_z=E_t,
            nu_xy=nu_lt,
            nu_xz=nu_lt,
            nu_yz=0.0,
            G_xy=G_lt,
            G_xz=G_lt,
            G_yz=0.0,
            density=density,
            failures=failures,
        )

    def get_E1(self) -> float:
        return self.E_l

    def get_E2(self) -> float:
        return self.E_t

    def get_nu12(self) -> float:
        return self.nu_lt

    def get_nu21(self) -> float:
        return self.get_nu12() * self.get_E2() / self.get_E1()

    def get_compliance(self) -> ndarray:
        compliance = np.zeros((6, 6))
        compliance[0, 0] = 1 / self.E_l
        compliance[1, 1] = 1 / self.E_t
        compliance[2, 2] = 1 / self.E_t

        compliance[1, 0] = compliance[0, 1] = -self.nu_lt / self.E_l
        compliance[2, 0] = compliance[0, 2] = -self.nu_lt / self.E_l
        compliance[2, 1] = compliance[1, 2] = -self.nu_tt / self.E_l

        compliance[3, 3] = 2 * (1.0 + self.nu_tt) / self.E_l
        compliance[4, 4] = 1 / self.G_lt
        compliance[5, 5] = 1 / self.G_lt

        return compliance



import pytest
from pymaterial.materials import OrthotropicMaterial


@pytest.mark.parametrize(
    "Es, nus, Gs, density, ndigits",
    [
        ((3.0, 2.0, 1.0), (0.3, 0.2, 0.1), (0.6, 0.5, 0.4), 1000, [3, 3, 3]),
    ],
)
def test_values(Es, nus, Gs, density, ndigits):
    material = OrthotropicMaterial(*Es, *nus, *Gs, density)
    compliance = material.get_compliance()
    assert len(compliance) == 6
    assert round(compliance[0, 0], ndigits[0]) == round(1 / Es[0], ndigits[0])
    assert round(compliance[0, 1], ndigits[1]) == round(-nus[0] / Es[0], ndigits[1])
    assert round(compliance[4, 4], ndigits[2]) == round(1 / Gs[1], ndigits[2])



from .material import Material, np, ndarray, Optional, List, IFailure


class OrthotropicMaterial(Material):
    def __init__(
        self,
        E_x: float,
        E_y: float,
        E_z: float,
        nu_xy: float,
        nu_xz: float,
        nu_yz: float,
        G_xy: float,
        G_xz: float,
        G_yz: float,
        density: float,
        failures: Optional[List[IFailure]] = None,
    ):
        self.E_x = E_x
        self.E_y = E_y
        self.E_z = E_z
        self.nu_xy = nu_xy
        self.nu_xz = nu_xz
        self.nu_yz = nu_yz
        self.G_xy = G_xy
        self.G_xz = G_xz
        self.G_yz = G_yz
        attr = dict(
            EX=E_x,
            EY=E_y,
            EZ=E_z,
            PRXY=nu_xy,
            PRXZ=nu_xz,
            PRYZ=nu_yz,
            GXY=G_xy,
            GXZ=G_xz,
            GYZ=G_yz,
            DENS=density,
        )
        super().__init__(attr, failures=failures)

    def get_compliance(self) -> ndarray:
        compliance = np.zeros((6, 6))
        compliance[0, 0] = 1 / self.E_x
        compliance[1, 1] = 1 / self.E_y
        compliance[2, 2] = 1 / self.E_z
        compliance[3, 3] = 1 / self.G_yz
        compliance[4, 4] = 1 / self.G_xz
        compliance[5, 5] = 1 / self.G_xy

        compliance[1, 0] = compliance[0, 1] = -self.nu_xy / self.E_x
        compliance[2, 0] = compliance[0, 2] = -self.nu_xz / self.E_x
        compliance[2, 1] = compliance[1, 2] = -self.nu_yz / self.E_y

        return compliance



import pytest
from pymaterial.materials import TransverselyIsotropicMaterial
from pymaterial.combis.clt import Ply, Stackup
import numpy as np
from math import ceil, log

material = TransverselyIsotropicMaterial(
    E_l=141000.0, E_t=9340.0, nu_lt=0.35, G_lt=4500.0, density=1.7e-9
)


def get_significance(number: float) -> int:
    significance = 0
    if number != 0.0:
        significance = abs(number)  # cant be negative
        significance = log(significance, 10)  #
        significance = abs(significance)  # will be negative
        significance = ceil(significance)  # rund_up
    return significance


@pytest.mark.parametrize(
    "plies, a_mat",
    [
        (
            [Ply(material, 1.0, 0)],
            np.array(
                [[142153.5, 3295.7, 0.0], [3295.7, 9416.4, 0.0], [0.0, 0.0, 4500.0]]
            ),
        ),
        (
            [Ply(material, 1.0, 45.0, degree=True)],
            np.array(
                [
                    [44040.4, 35040.4, 33184.3],
                    [35040.4, 44040.4, 33184.3],
                    [33184.3, 33184.3, 36244.6],
                ]
            ),
        ),
        (
            [Ply(material, 1.0, -45.0, degree=True)],
            np.array(
                [
                    [44040.4, 35040.4, -33184.3],
                    [35040.4, 44040.4, -33184.3],
                    [-33184.3, -33184.3, 36244.6],
                ]
            ),
        ),
    ],
)
def test_abd_a(plies: list, a_mat):
    abd = Stackup(plies).get_abd()
    for i in range(3):
        for j in range(3):
            assert round(abd[i, j], 1) == a_mat[i, j]


@pytest.mark.parametrize(
    "plies, b_mat",
    [
        (
            [Ply(material, 1.0, 0)],
            np.array([[11846.1, 274.6, 0.0], [274.6, 784.7, 0.0], [0.0, 0.0, 375.0]]),
        ),
        (
            [Ply(material, 1.0, 45.0, degree=True)],
            np.array(
                [
                    [3670.0, 2920.0, 2765.4],
                    [2920.0, 3670.0, 2765.4],
                    [2765.4, 2765.4, 3020.4],
                ]
            ),
        ),
        (
            [Ply(material, 1.0, -45.0, degree=True)],
            np.array(
                [
                    [3670.0, 2920.0, -2765.4],
                    [2920.0, 3670.0, -2765.4],
                    [-2765.4, -2765.4, 3020.4],
                ]
            ),
        ),
    ],
)
def test_abd_b(plies: list, b_mat):
    abd = Stackup(plies).get_abd()
    for i in range(3):
        for j in range(3):
            assert round(abd[i + 3, j + 3], 1) == b_mat[i, j]


@pytest.mark.parametrize(
    "plies, loading, deformations",
    [
        (
            [Ply(material, 1.0, 45.0, degree=True)],
            [1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            (0.000083, -0.000028, -0.00005, 0.0, 0.0, 0.0),
        ),
        (
            [Ply(material, 1.0, -45.0, degree=True)],
            [1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            (0.000083, -0.000028, 0.00005, 0.0, 0.0, 0.0),
        ),
    ],
)
def test_deformations(plies: list, loading, deformations):
    deform = Stackup(plies).apply_load(np.array(loading))
    assert len(deform) == len(deformations)
    for i in range(len(deformations)):
        significance = get_significance(deformations[i])
        assert round(deform[i], significance + 1) == deformations[i]


@pytest.mark.parametrize(
    "plies, loading, strains",
    [
        (
            [Ply(material, 1.0, 45.0, degree=True)],
            [1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [([2.3e-06, 5.2e-05, -1.1e-04], [2.3e-06, 5.2e-05, -1.1e-04])],
        ),
        (
            [Ply(material, 1.0, -45.0, degree=True)],
            [1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [([2.3e-06, 5.2e-05, 1.1e-04], [2.3e-06, 5.2e-05, 1.1e-04])],
        ),
    ],
)
def test_strains(plies: list, loading, strains):
    stackup = Stackup(plies)
    deform = stackup.apply_load(np.array(loading))
    layer_strains = stackup.get_strains(deform)
    assert len(layer_strains) == len(strains)
    for i in range(len(strains)):
        # bottom
        stress = strains[i][0]
        for j in range(len(stress)):
            significance = get_significance(stress[j])
            assert round(layer_strains[i][0][j], significance + 1) == stress[j]

        # top
        stress = strains[i][1]
        for j in range(len(stress)):
            significance = get_significance(stress[j])
            assert round(layer_strains[i][1][j], significance + 1) == stress[j]


@pytest.mark.parametrize(
    "plies, loading, stresses",
    [
        (
            [Ply(material, 1.0, 45.0, degree=True)],
            [1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [([0.5, 0.5, -0.5], [0.5, 0.5, -0.5])],
        ),
        (
            [Ply(material, 1.0, -45.0, degree=True)],
            [1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [([0.5, 0.5, 0.5], [0.5, 0.5, 0.5])],
        ),
    ],
)
def test_stresses(plies: list, loading, stresses):
    stackup = Stackup(plies)
    deform = stackup.apply_load(np.array(loading))
    layer_strains = stackup.get_strains(deform)
    layer_stresses = stackup.get_stresses(layer_strains)
    assert len(layer_stresses) == len(stresses)
    for i in range(len(stresses)):
        # bottom
        stress = stresses[i][0]
        for j in range(len(stress)):
            significance = get_significance(stress[j])
            assert round(layer_stresses[i][0][j], significance + 1) == stress[j]

        # top
        stress = stresses[i][1]
        for j in range(len(stress)):
            significance = get_significance(stress[j])
            assert round(layer_stresses[i][1][j], significance + 1) == stress[j]



import numpy as np
from pymaterial.failures import IFailure
from typing import Optional, List, Union
from numpy import ndarray


class Material(IFailure):
    def __init__(self, attr: dict, failures: Optional[List[IFailure]] = None):
        """
        Parameters
        ----------
        attr : dict
        failures : List[IFailure], optional
        """
        self.attr = attr
        if failures is None:
            failures = []
        self.failures = failures

    def get_compliance(self) -> ndarray:
        raise NotImplementedError

    def get_stiffness(self) -> ndarray:
        return np.linalg.inv(self.get_compliance())

    def get_density(self) -> float:
        return self.attr.get("DENS")

    def get_failures(self) -> List[IFailure]:
        """
        Returns
        -------
        List[IFailure]
            List of possible material failures
        """
        return self.failures

    def get_failure(
        self,
        stresses: Optional[Union[List[float], ndarray]] = None,
        strains: Optional[Union[List[float], ndarray]] = None,
        temperature: Optional[float] = None,
    ) -> dict:
        """
        returns
        {"max_stress": 1.0, "cuntze": 0.5}
        """
        result = dict()
        for failure in self.failures:
            result.update(failure.get_failure(stresses, strains, temperature))
        return result

    def get_plane_stress_stiffness(self):
        """
        Get stiffness tensor for plane stress
        Returns
        -------
        """
        elems = [0, 1, 5]  # ignore the s_zz, s_xy and s_yz row and column
        stiffness = self.get_stiffness()
        return stiffness[elems][:, elems]

    def get_plane_strain_stiffness(self):
        """
        et stiffness tensor for plane strain
        Returns
        -------
        """
        elems = [0, 1, 5]  # ignore the s_zz, s_xy and s_yz row and column
        return np.linalg.inv(self.get_compliance()[elems][:, elems])



import pytest
from pymaterial.failures import VonMisesFailure


@pytest.mark.parametrize(
    "strength, stress, result, ndigits",
    [
        (200, [0, 0, 0], 0.0, 2),
        (200, [100, 0, 0], 0.5, 1),
        (200, [200, 0, 0], 1.0, 1),
        (200, [100, 0, 0, 0, 0, 0], 0.5, 1),
        (200, [200, 0, 0, 0, 0, 0], 1.0, 1),
    ],
)
def test_values(strength, stress, result, ndigits):
    failure = VonMisesFailure(strength)
    res = failure.get_failure(stress)
    assert round(res["mises"], ndigits) == result


@pytest.mark.parametrize(
    "val",
    [
        (0),
        (-0.001),
        (-1),
    ],
)
def test_negative_strength_exeption(val):
    with pytest.raises(ValueError):
        VonMisesFailure(val)


def test_none_stress():
    failure = VonMisesFailure(1)
    with pytest.raises(ValueError):
        failure.get_failure()


@pytest.mark.parametrize(
    "stresses",
    [
        ([1]),
        ([1, 2]),
        ([1, 2, 3, 4]),
        ([1, 2, 3, 4, 5]),
        ([1, 2, 3, 4, 5, 6, 7]),
    ],
)
def test_incorrect_stress_lenght(stresses):
    failure = VonMisesFailure(1)
    with pytest.raises(ValueError):
        failure.get_failure(stresses)



from .material import Material, np, ndarray


class IsotropicMaterial(Material):
    def __init__(self, Em: float, nu: float, density: float, **kwargs):
        """
        Parameters
        ----------
        Em
        nu
        density
        kwargs
        """
        attr = dict(EX=Em, PRXY=nu, DENS=density)
        self.Em = Em
        self.nu = nu
        super().__init__(attr, **kwargs)

    def get_compliance(self) -> ndarray:
        compliance = np.zeros((6, 6))
        compliance[0, 0] = 1 / self.Em
        compliance[1, 1] = 1 / self.Em
        compliance[2, 2] = 1 / self.Em

        compliance[1, 0] = compliance[0, 1] = -self.nu / self.Em
        compliance[2, 0] = compliance[0, 2] = -self.nu / self.Em
        compliance[2, 1] = compliance[1, 2] = -self.nu / self.Em

        compliance[3, 3] = 2 * (1 + self.nu) / self.Em
        compliance[4, 4] = 2 * (1 + self.nu) / self.Em
        compliance[5, 5] = 2 * (1 + self.nu) / self.Em

        return compliance



from .material import Material, ndarray, np


class AnisotropicMaterial(Material):
    def __init__(self, stiffness: ndarray, density: float, **kwargs):
        """
        Parameters
        ----------
        stiffness
        density
        kwargs
        """
        self.stiffness = stiffness
        super().__init__(dict(DENS=density), **kwargs)

    def get_stiffness(self) -> ndarray:
        return self.stiffness

    def get_compliance(self) -> ndarray:
        return np.linalg.inv(self.get_stiffness())



import numpy as np
from pymaterial.materials import Material


class Ply:
    def __init__(
        self, material: Material, thickness: float, rotation=0.0, degree=False
    ):
        """
        Ply or Lamina
        Parameters
        ----------
        material : Material
            material of the ply
        thickness : float
            thickness of the ply
        rotation : float, optional
            rotation of the ply in [rad] or [deg] when degree is set True
        degree : bool, optional
            changes the measurement system of rotation, default is False
        References
        ----------
        .. [1] J. Ashton and J.M. Whitney, "Theory of Laminated Plates",
           Technomic, vol. 4, 1970
        """
        self.material = material
        self.thickness = thickness
        if degree:
            rotation = rotation * np.pi / 180.0
        self.rotation = rotation
        self.plane_strain = self.material.get_plane_strain_stiffness()

    def rotate(self, rad, degree=False) -> "Ply":
        """
        Rotate Ply
        Parameters
        ----------
        angle
            angle in [rad] or [deg] when degree is set True
        degree : bool, optional
            changes the measurement system of angle, default is False
        Returns
        -------
        Stackup
            new instance of the rotated Stackup
        Examples
        --------
        >>> ply = Ply(material, 1.0, 0.0)
        >>> rot_ply = ply.rotate(90, degree=True)
        Be aware, that the ply is left untouched.
        To change this, use:
        >>> ply = ply.rotate(90, degree=True)
        instead.
        """
        if degree:
            rad = rad * np.pi / 180.0
        return Ply(self.material, self.thickness, self.rotation + rad)

    def get_stiffness(self, local=False) -> np.ndarray:
        """Returns the Stiffness or Q-Tensor (always 3x3)
        Parameters
        ----------
        local : bool, optional
            when true return the stiffness in the local coordinate system
        Returns
        -------
        array:
            the stiffness tensor in rotated or local coordinate system
        """
        if local:
            return self.plane_strain
        c = np.cos(self.rotation)
        s = np.sin(self.rotation)
        t_sigma = np.array(
            [
                [c**2, s**2, -2 * s * c],
                [s**2, c**2, 2 * s * c],
                [s * c, -s * c, c**2 - s**2],
            ]
        )

        t_sigma_t = np.array(
            [
                [c**2, s**2, s * c],
                [s**2, c**2, -s * c],
                [-2 * s * c, 2 * s * c, c**2 - s**2],
            ]
        )
        stiffness_rot = np.matmul(t_sigma, np.matmul(self.plane_strain, t_sigma_t))
        return stiffness_rot

    def get_material(self) -> Material:
        """
        Material of the ply
        Returns
        -------
        Material
            Material of the ply.
        """
        return self.material

    def get_thickness(self) -> float:
        """
        Thickness of the ply
        Returns
        -------
        float
            Thickness of the ply.
        """
        return self.thickness

    def get_rotation(self, degree=False) -> float:
        """
        Rotation of the ply.
        Parameters
        ----------
        degree : bool, optional
            changes the measurement system of the return value, default is False
        Returns
        -------
        float:
            Rotation angle of the ply.
            By default the Unit is in [rad], but can be [deg] if degree is set to True.
        """
        if degree:
            return self.rotation * 180.0 / np.pi
        return self.rotation

    def get_local_stress(self, stress: np.ndarray) -> np.ndarray:
        """
        Stress in ply coordinate system.
        Parameters
        ----------
        stress : array
            stress tensor in Voigt notation / [sig_11, sig_22, sig_12]
        Returns
        -------
        array
            2d stress tensor in Voigt notation
        """
        angle = self.rotation
        c = np.cos(angle)
        s = np.sin(angle)
        t_sigma_inv = np.array(
            [
                [c**2, s**2, 2 * c * s],
                [s**2, c**2, -2 * c * s],
                [-c * s, c * s, c**2 - s**2],
            ]
        )
        return np.ravel(t_sigma_inv.dot(stress))

    def get_local_strain(self, strain: np.ndarray) -> np.ndarray:
        """
        Strain in ply coordinate system.
        Parameters
        ----------
        strain : array
            strain tensor in Voigt notation / [eps_11, eps_22, gamma_12]
        Returns
        -------
        array
            2d strain tensor in Voigt notation
        """
        angle = self.rotation
        c = np.cos(angle)
        s = np.sin(angle)
        t_epsilon_inv = np.array(
            [
                [c**2, s**2, c * s],
                [s**2, c**2, -c * s],
                [-2 * c * s, 2 * c * s, c**2 - s**2],
            ]
        )
        return np.ravel(t_epsilon_inv.dot(strain))
